#include <iostream>
#include <thread>         // std::thread
#include <assert.h>
#include <chrono>  // for high_resolution_clock

int global;

thread_local int __DynamicScope__tls_variable;
template <class T>
class DynamicScope {
    int * tls_guard;
    T  instance;
    T* next;

    void __check_tls() {
        assert(tls_guard == & __DynamicScope__tls_variable
               && "You have created a DynamicScope variable without thread_local");
    }

public:
    template <class... T2>
    DynamicScope<T>(T2... rest)
        : instance(rest...),
          next(&instance),
          /* With thread_local, the constructor is executed  once for every thread */
          //root_thread(std::this_thread::get_id())
          tls_guard(&__DynamicScope__tls_variable)
        { /* empty */ }

    T& operator*()  {
        __check_tls();
        return *next;
    }
    T* operator->() {
        __check_tls();
        return next;
    }

    // Create a new binding at the given DynamicScope Point
    class BindPointer {
        T* old_next;
        DynamicScope<T> &head;
    public:
        BindPointer(DynamicScope<T>& head, T* instance)
            : head(head), old_next(head.next) {
            head.__check_tls();
            head.next = instance;
        }

        ~BindPointer() {
            head.__check_tls();
            head.next = old_next;
        }
    };

    class BindInstance : protected BindPointer{
        T instance;
    public:
        template <class... T2>
        BindInstance(DynamicScope<T>& head, T2... rest)
            : BindPointer(head, &instance), instance(rest...) {}
    };

};

thread_local DynamicScope<int> G(0);


void bar() {
    std::cout << "bar " << *G << std::endl;
}

void foo(int x) {
    std::cout << "foo(" << x << ")=" << *G << std::endl;
    DynamicScope<int>::BindInstance _(G, x);
    bar();
}

void timeit() {
    // Record start time
    auto start = std::chrono::high_resolution_clock::now();

    for (unsigned i = 0; i< (1<<30); i++) {
        global += *G;
    }

    // Record end time
    auto finish = std::chrono::high_resolution_clock::now();

    std::chrono::duration<double> elapsed = finish - start;
    std::cout << "Elapsed time for 1<<30 deferences: " << (elapsed.count() * 1000) << " ms\n";
}

int main() {
    std::cout << "main1 " << *G << std::endl;

    foo(1);
    {
        decltype(G)::BindInstance _(G, 33);
        foo(2);

        std::thread first(foo, 3);

        for (int i= 0; i < 1000; i++)
            timeit();



        first.join();
    }
    std::cout << "main2 " << *G << std::endl;
}
